/*
 * client_test.c
 * a test harness for rpcap_client
 * Copyright (C) S. Krishnan <sri_krishnan@yahoo.com>, 2002
 */


/*
 * This file is part of the RPCAP Remote Packet Capture System.
 *
 * RPCAP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */


#include <stdio.h>
#include<stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
/* #include <pcap/pcap.h> */
#include <signal.h>
#include </usr/local/include/pcap.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "pcap-int.h"

/*
 * This file is also intended to serve as a quick introduction to using rpcap.
 * I have provided comments wherever necessary.  Where there are no comments, you can
 * safely assume that rpcap follows the libpcap API faithfully.
 */

int rpcap_end_capture_handler(void);

pcap_t *remhandle, localhdl;
char errbuf[256];
FILE *savedump;

pcap_dumper_t *dumpfile_handle;

/* the various data link layer types that libpcap lists in net/bpf.h */
char *datalink_type[] =
{
	"None -  no link-layer encapsulation",
	"Ethernet (10Mb)",
	"Experimental Ethernet (3Mb)",
	"Amateur Radio AX.25",
	"Proteon ProNET Token Ring",
	"Chaos",
	"IEEE 802 Networks",
	"ARCNET",
	"Serial Line IP",
	"Point-to-point Protocol",
	"FDDI"
};


/*
 * This is the callback function that is invoked from pcap_dispatch and pcap_loop;
 *	it has an identical signature and usage to the libpcap version
 */

void pkt_handler(u_char *user, struct pcap_pkthdr *phdr, const u_char *pkt_data)
{
	int i;
	unsigned int len;
	const u_char *pktdata_ptr=pkt_data;
	
	len = phdr->caplen;
	printf("Pkt Handler called\n");
	printf("Capture length = %u\n", len);
	printf("Dumping capture data\n");
	
	for(i=0; i < len; i++)
		printf("%d ", *pktdata_ptr++);
		
	pcap_dump((u_char *)dumpfile_handle, phdr, pkt_data);
}

void sigint_hdlr()
{
	/*
	 * this function is invoked if the user hits a Ctrl-C to terminate the program.
	 * it shuts down the capture mechanism in an orderly fashion, flushes the
	 * file buffers to disk and then exits
	 */

	int result;
	struct pcap_stat ps;
	/* close the dump file */
	pcap_dump_close(dumpfile_handle);
	result = pcap_stats(remhandle, &ps);
	printf("pcap_stats returned recd. packets %u\t dropped packets %u\n", ps.ps_recv, ps.ps_drop);
	pcap_close(remhandle);
	/* stop the capture */
	rpcap_end_capture_handler();
	/* flush buffers to disk */
	exit(0);
}



int main()
{
	u_char inbuf[66000];
	int result;

	char *devname;
	bpf_u_int32 netp, maskp;
	struct in_addr remote_addr;
	struct in_addr mask;
	char *filter_string;                    // = "ip proto 6";  /* this sets a filter for tcp packets */
	struct bpf_program fp;

	struct sigaction action, old_action;

	char *eb;
	
	/*
	 * init_rpcap is the first command to be executed when using rpcap.
	 * It does the following:
	 * 1. Initialization of the rpc mechanism: this is done by passing
	 *    the name of the remote host on which the capture server is
	 *		running ("ambipapa" in this example, but please substitute
	 *    the actual host name of the remote computer for your particular
	 *    case, as specified in your /etc/hosts or its moral equivalent thereof).
	 *    For example, if you were capturing packets off a remote system 
	 *    called fred, you would pass "fred" as the first parameter
	 *    to init_rpcap.
	 * 2. Passes the local client's host IP address and capture data
	 *    port number to both the client and remote systems.  This
	 *    ip address has to be supplied by the calling client program.
	 *    I chose this method rather than a gethostbyname kind of schema
	 *    because it is far more flexible and safe.  This information
	 *    is used to set up the capture data uplink between the remote
	 *    target machine and the local client.
	 *
	 * For example, assuming that you want to sniff network traffic on
	 * the remote host fred (the server), whose IP address is 192.168.1.10/24, 
	 * and the client machine (which will receive and process the captured
	 * data), called, say, wilma, has an IP address of 192.168.1.4.  We
	 * will use the tcp port 21373 for receipt of the capture stream.  Then
	 * init_rpcap will be invoked as follows:
	 *
	 * init_rpcap("fred", "192.168.1.4", 21373);
	 *            ------   -----------   -----
	 *            Target   wilma's IP    wilma's
	 *            machine   address      port no.
	 */


	/* install the signal handler to flush the capture save file buffer */
	action.sa_handler  = sigint_hdlr;
	action.sa_flags = 0;
	if(sigaction(SIGINT, &action, &old_action)==-1)
	{
		perror("Could not install termination handler for client_test");
		fprintf(stderr, "You may not be able to save the capture file\n");
	}
	
	if((result = init_rpcap("localhost", /* "ambipapa" */ "192.168.2.1", 21373))==0)
	{
		perror("main: error initializing server");
		exit(0);
	}
	
	/*
	 * the following are all standard libpcap functions, which you can
	 * look up in man (3) pcap and the large body of documentation
	 * that is available for it
	 */
		
	devname=pcap_lookupdev(errbuf);

	if(devname==NULL)
	{
		printf("rpcap_lookpdev returned NULL\n");
		sleep(10);
		exit(0);
	}
	printf("Looked up device %s\n", devname);
	
	pcap_lookupnet(devname, &netp, &maskp, errbuf);
	remote_addr.s_addr = netp;
	mask.s_addr = maskp;
	printf("rpcap_lookupnet returned net address %s\n", inet_ntoa(remote_addr));
	printf("rpcap_lookupnet returned net mask %s\n", inet_ntoa(mask));

	remhandle  = pcap_open_live("eth0", 576, 0, 10, errbuf);
	if(remhandle == NULL)
	{
		printf("remhandle is NULL\n");
		printf("Returned error is %s\n", errbuf);
		exit(0);
	}

	printf ("Returned from open_live\n");
	//	memcpy(&localhdl, remhandle, sizeof(pcap_t));
	
	result=pcap_compile(remhandle, &fp, filter_string, 0, maskp);
	printf("pcap_compile returned %d\n", result);

	/* we set the filter for protocol #6, which is tcp, in this example */
	/*
	result= pcap_setfilter(remhandle, &fp);
	printf("pcap_setfilter returned %d\n", result);
	*/

	eb = pcap_geterr(remhandle);

	fprintf(stderr, "EB: %s\n", eb);
	
	dumpfile_handle = pcap_dump_open(&localhdl, "remdump");
	
	result = pcap_snapshot(remhandle);
	printf("Capture snapshot length = %d\n", result);

	result = pcap_datalink(remhandle);
	if(result >=0 && result <=10)
	  printf("Datalink layer type is %s\n", datalink_type[result]);
	else
	  printf("Datalink layer type reported is %d - interpret it as you will\n", result);

	

	printf("starting pcap_loop\n");
	result = pcap_loop(remhandle, 0, (pcap_handler)pkt_handler, inbuf);
	/* pcap_dispatch under Linux gets only a single packet */
	
	printf("CLIENT_TEST: Pcap_dispatch / pcap_loop returned %d\n", result);
	

}

	
	
