/*******************************************************************************************/
/*                                file rpcap.c                                             */
/*                                                                                         */
/*           Copyright (C) S. Krishnan <sri_krishnan@yahoo.com>, 2002                      */
/*                                                                                         */
/*     Server file for the rpcap Remote Packet Capture System                              */
/*     Contains the implementation of the wrapper functions for the rpcap RPC calls        */
/*		 defined in rpcap.h                                                                  */
/*     language: C                                                                         */
/*     Compiler: gcc                                                                       */
/*     Dependencies: Libpcap, RPC libraries                                                */
/*******************************************************************************************/
 
 
/*
 * This file is part of the RPCAP Remote Packet Capture System.
 *
 * RPCAP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include </usr/local/include/pcap.h> 
#include <fcntl.h>
#include <errno.h>
#include <rpc/rpc.h>


#ifdef SYS_LIBPCAP
    #include <pcap/pcap.h>
#elif defined LOCAL_LIBPCAP
    #include </usr/local/include/pcap.h>
#endif  /* LIBPCAP includes */

#include "pcap-int.h"
#include "rpcap.h"

/*------------------------- Global Declarations ---------------------------------------*/

struct sockaddr_in remote_conn;
int conn_flag = 0;
pcap_t *la;

char errbuf[PCAP_ERRBUF_SIZE];

typedef int (*capture_function)(pcap_t *p, int cnt, pcap_handler callback, u_char *user);
capture_function capfunc_array[] = {pcap_dispatch, pcap_loop};

typedef enum cap_func_index
{
	PCAP_DISPATCH=0,
	PCAP_LOOP,
} CAP_FUNC_INDEX;

struct capture_params
{
		CAP_FUNC_INDEX cap_func_selector;
		pcap_t cap_interface;
		int cap_count;
		pcap_handler callback;
		struct cl_addr cap_client;
};

struct capture_params dispatch_params;
struct capture_params *cparams;

pthread_t cap_thread;

void *start_capture(void *cap_args);
int client_dump(int capture_retval, u_long client_ip);
void sigpipe_hdlr(int);

pcap_t tmphdl;
/*-------------------------------------------------------------------------------------*/



void packet_pump(u_char *u_buff, struct pcap_pkthdr *cap_header, const u_char *pkt_data)
{
	static int pkt_fd;
	fd_set pkt_set;
	struct timeval tv;
	struct sockaddr_in *client_conn;
	unsigned cap;
	int wrt_cnt=0, nbytes=0, sel_val;
	u_short hdrlen=0;
	long timebuf;
	
	tv.tv_sec = 3;
	tv.tv_usec = 0;

	if(conn_flag==2)
	{
		fprintf(stderr, "\nRPCAP_SVC: Closing the capture data link socket\n");
		close(pkt_fd);
		goto Quit;
	}
	
	if(conn_flag==0)
	{
		if((pkt_fd=socket(AF_INET, SOCK_STREAM, 0)) <0)
		{
			perror("Error creating data uplink socket!\n");
			goto Quit;
		}
		
		conn_flag=1;

		client_conn = (struct sockaddr_in *)u_buff;
		if(connect(pkt_fd, (struct sockaddr *)client_conn, sizeof(remote_conn)) <0)
		{
			if(errno == EINPROGRESS)
				printf("PACKET_PUMP:: Nonblocking socket - will connect ASAP\n");        /* Legacy crap - remove ASAP */
			else
				perror("Error connecting to the data receiver");
		}
	}
	
	FD_ZERO(&pkt_set);
	FD_SET(pkt_fd, &pkt_set);
	sel_val = select((pkt_fd+1), NULL, &pkt_set, NULL, NULL);
	if(sel_val==-1)
	{
		fprintf(stderr, "RPCAP Server::packet_pump - SELECT error\n");
		goto Quit;
	}
	
	hdrlen = htons(sizeof(cap_header));
	write(pkt_fd, &hdrlen, sizeof(u_short));
	timebuf = htonl(cap_header->ts.tv_sec);
	write(pkt_fd, &timebuf, sizeof(long));
	timebuf = htonl(cap_header->ts.tv_usec);
	write(pkt_fd, &timebuf, sizeof(long));
	cap = htonl(cap_header->caplen);
	write(pkt_fd, &cap, sizeof(unsigned));
	cap = htonl(cap_header->len);
	write(pkt_fd, &cap, sizeof(unsigned));

	while(nbytes <cap_header->caplen)
	{
		if((wrt_cnt=write(pkt_fd, pkt_data, (cap_header->caplen)))==-1)
			perror("Write error to RPCAP client socket");
		else
		{
			nbytes += wrt_cnt;
#ifdef DEBUG
			fprintf(stderr, "Wrote %d characters\n", wrt_cnt);
#endif
		}
	}
	  
	Quit:	
}



adapter_ptr * rpcap_open_live_1_svc(RPCAP_OPEN_LIVE_PARAMS *open_params, struct svc_req *rqst)
{
	static pcap_t *local_adapter;
	static adapter_handle remote_adapter;
	static adapter_ptr result;
	
	if(open_params->device == NULL)
	{
		fprintf(stderr, "pcap_open_live: null device\n");
		return NULL;
	}
	local_adapter = pcap_open_live(open_params->device, open_params->snaplen, open_params->promisc,
		open_params->to_ms, errbuf);
	memset(&remote_adapter, 0, sizeof(adapter_handle));
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val= local_adapter;
	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);

	result = &remote_adapter;
	/* la = local_adapter; */
	return &result; 
}



RPCAP_FN_RTN_PTR * rpcap_lookupdev_1_svc(void *no_param, struct svc_req *rqst)
{
	static rpcap_func_rtn lookupdev_ret;
	static RPCAP_FN_RTN_PTR lookupdev_rtn_ptr;
	static char *device;
	if((device = pcap_lookupdev(errbuf))==NULL)
	{
		lookupdev_ret.ebuf = errbuf;
		lookupdev_ret.retval = -1;
	}
	else
	{
		lookupdev_ret.ebuf = device;
		lookupdev_ret.retval = 0;
	}
	lookupdev_rtn_ptr = &lookupdev_ret;
	return &lookupdev_rtn_ptr;
}



LOOKUPNET_RESULT_PTR *rpcap_lookupnet_1_svc(char **dev_ptr, struct svc_req *rqst)
{
	static RPCAP_LOOKUPNET_RESULT result;
	static LOOKUPNET_RESULT_PTR result_ptr;
	bpf_u_int32 netp, maskp;
	int lookup;
	
	if(*dev_ptr==NULL)
	{
		fprintf(stderr,"rpcap_lookupnet: passed parameter is NULL\n");
		result.lookupnet_ret_val = -1;
		result.netp = -1;
		result.maskp = -1;
		result_ptr=&result;
		return &result_ptr;
	}
	lookup=pcap_lookupnet(*dev_ptr, &netp, &maskp, errbuf);
	result.lookupnet_ret_val = lookup;
	result.netp = netp;
	result.maskp = maskp;
	result.ebuf = errbuf;
	result_ptr = &result;
	
	return &result_ptr;
}



int * rpcap_capture_1_svc(RPCAP_CAPTURE_PARAMS* rpcap_cap_dispatch, struct svc_req *rqst)
{
	/*
	 * we get all the capture parameters passed to us here except the callback function
	 * address, which of course has to be local.  An additional parameter passed to us here
	 * from the client is the TCP port number on which the client will be listening for the
	 * capture packets which the callback will pass on, which is then passed to the callback
	 * service function (the "packet pump") for servicing the client
	 */

	static int result;

	int *rptr = &result;
	struct cl_addr *client_channel;
	int cap_thread_id;

#ifdef DEBUG
	fprintf(stderr, "Starting pcap_capture\n");
#endif
	dispatch_params.cap_func_selector = rpcap_cap_dispatch->cap_func_sel;
	memcpy(&dispatch_params.cap_interface, (pcap_t *)rpcap_cap_dispatch->p->adapter_pcap.adapter_pcap_val,
		sizeof(pcap_t));
	dispatch_params.cap_count = rpcap_cap_dispatch->cnt;
	dispatch_params.callback = (pcap_handler)packet_pump;
	memcpy(&dispatch_params.cap_client,  &rpcap_cap_dispatch->client_address, sizeof(struct cl_addr));
	
	if((cap_thread_id=pthread_create(&cap_thread, NULL, start_capture, (void *)&dispatch_params)))
	{
		perror("Capture thread creation error");
		result = -1;
		return &result;
	}
	result=1;
	return &result;
}



void *start_capture(void *cap_args)
{
	static int retval;
	/* struct capture_params *cparams; - had to make this global so that the thread kill function could get at it - sigh... */
	struct sigaction action, old_action;
	struct sockaddr_in client_conn;
	
	cparams = (struct capture_params *)cap_args;
	
	client_conn.sin_family = AF_INET;
	/* note that port no is in host order while the client ip address is in network byte order: fix this??  */
	client_conn.sin_port = htons(cparams->cap_client.cl_port);
	client_conn.sin_addr.s_addr = cparams->cap_client.cl_ip;

	/* set up the signal handler for SIGPIPE */
	action.sa_handler  = sigpipe_hdlr;
	action.sa_flags = 0;
	if(sigaction(SIGPIPE, &action, &old_action)==-1)
	{
		perror("RPCAP::start_capture - Error on SIGACTION call");
		retval=-1;
		return (int *)&retval;
	}

	retval = capfunc_array[cparams->cap_func_selector]
		(&cparams->cap_interface, cparams->cap_count, cparams->callback, (u_char *)&client_conn);

	capture_cleanup(retval);
	return &retval;
}



BPF_PRG_PTR *rpcap_compile_1_svc(RPCAP_COMPILE_PARAMS *compile_params, struct svc_req *rqst)
{
	/* we just have to compile the string we receive from the client */

	static bpf_prg rpcap_filter;
	static BPF_PRG_PTR filter_ptr;
	int result;
	static struct bpf_program fp;
	char *str;
	
	if(compile_params->p ==NULL || compile_params->str ==NULL)
		fprintf(stderr, "rpcap_compile: a parameter had NULL value\n");

	if(!strcmp(compile_params->str, "NULL"))
		str = NULL;
	else
		str = compile_params->str;
	result=pcap_compile((pcap_t *)compile_params->p->adapter_pcap.adapter_pcap_val, &fp,
		 str, compile_params->optimize, compile_params->netmask);
	if (result == -1)
	{
		fprintf(stderr, "PCAP_COMPILE ERROR\n");
		filter_ptr=NULL;                         /* NULL return to client code means error */
	}
			
	else
	{
#ifdef DEBUG
		fprint(stderr, "PCAP_COMPILE returned %d\n", result);
#endif
		rpcap_filter.bpf_prgrm.bpf_prgrm_len = sizeof(struct bpf_program);
		(struct bpf_program *)rpcap_filter.bpf_prgrm.bpf_prgrm_val=&fp;
		filter_ptr = &rpcap_filter;
	}
	return &filter_ptr;
}


int * rpcap_setfilter_1_svc(RPCAP_SETFILTER_PARAMS *filter_params, struct svc_req *rqst)
{
	static int result;
	if(filter_params->p==NULL || filter_params->fp == NULL)
	{
		perror("rpcap_setfilter: a calling parameter was NULL\n");
		result=-1;
		return &result;
	}
	result=pcap_setfilter((pcap_t *)filter_params->p->adapter_pcap.adapter_pcap_val,
		(struct bpf_program *)filter_params->fp->bpf_prgrm.bpf_prgrm_val);
	return &result;
}


int * rpcap_datalink_1_svc(adapter_handle *p, struct svc_req *rqst)
{
	static int result;
	if(p==NULL)
	{
		fprintf(stderr, "rpcap_datalink: adapter handle NUll value\n");
		result=-1;
		return &result;
	}
	result=pcap_datalink((pcap_t *)p->adapter_pcap.adapter_pcap_val);
	return &result;
}
	

int * rpcap_snapshot_1_svc(adapter_handle *p, struct svc_req *rqst)
{
	static int result;
	if(p==NULL)
	{
		fprintf(stderr, "rpcap_snpashot: adapter handle NULL value\n");
		result=-1;
		return &result;
	}
	result=pcap_snapshot((pcap_t *)p->adapter_pcap.adapter_pcap_val);
	return &result;
}



int * rpcap_close_1_svc(adapter_handle *p, struct svc_req *rqst)
{
	static int result;
	if(p==NULL)
	{
		fprintf(stderr, "rpcap_close: adapter handle NULL value\n");
		result=-1;
		return &result;
	}
	pcap_close((pcap_t *)p->adapter_pcap.adapter_pcap_val);
	result=1;
	return &result;
}



STATS_WRAPPER_PTR * rpcap_stats_1_svc(adapter_handle *p, struct svc_req *rqst)
{
	pcap_t *kludged_handle;
	static RPCAP_STATS_WRAPPER rpcap_statistics;
	static STATS_WRAPPER_PTR stats_ptr;
	static struct pcap_stat ps;
	
	if(&dispatch_params.cap_interface==NULL)
	{
		fprintf(stderr, "rpcap_close: adapter handle NULL value\n");
		rpcap_statistics.stats_retval=-1;
		stats_ptr=&rpcap_statistics;
		return &stats_ptr;
	}
	
	/*
	 * There is a massive kludge here to get the statistics.  Actually, libpcap stores packet
	 * capture statistics as a part of the pcap_t structure that acts as a handle to the
	 * network interface.  This creates a problem with our implementation of rpcap, since
	 * pointers are one-way in the RPC dispensation.  Once we pass the adapter handle to
	 * rpcap_capture_1_svc, only the local copy of the pcap_t * handle get updated with the
	 * stats.  However, when we call rpcap_stats_1_svc with the adapter handle as a
	 * parameter, we are passing it our copy of the original adapter handle, with md.stat
	 * values as they were *before* capture started, so that we would not get the stats at all.
	 * This implementation is a temporary kludge to get stats, by passing the "used" handle
	 * from start_capture, which is anyway a global (dispatch_params).
	 * To be fixed on the next visit.
	 */

	rpcap_statistics.stats_retval = pcap_stats(&dispatch_params.cap_interface, &ps);
	rpcap_statistics.pcap_stats.pcap_stats_len = sizeof(struct pcap_stat);
	(struct pcap_stat *)rpcap_statistics.pcap_stats.pcap_stats_val = &ps;
	stats_ptr=&rpcap_statistics;
	return &stats_ptr;
}
	


RPCAP_FN_RTN_PTR * rpcap_geterr_1_svc(adapter_handle *p, struct svc_req *rqst)
{
	/* Note: we're using the same kludge here as in pcap_stats: FIXME asap */
	static rpcap_func_rtn current_error;
	static RPCAP_FN_RTN_PTR current_error_ptr;
	
	static char ebuf[256];
	ebuf[0] = '\0';

	/* fprintf(stderr, "PCAP_GETERR INVOKED\n"); */

	strcpy(ebuf, pcap_geterr(&dispatch_params.cap_interface));
	if (strlen(ebuf) == 0)	
		strcpy(ebuf, "No error reported");

	/* fprintf(stderr, "PCAP_GETERR: EBUF = %s\n", ebuf); */
	current_error.ebuf = ebuf;
	current_error.retval = 0;
	
	current_error_ptr = &current_error;
	return &current_error_ptr;
} 


/*
char ** rpcap_strerror_1_svc(int *errnumber, struct svc_req *cl)
{
	static char *errdesc;

	errdesc = strerror(*errnumber);
	return &errdesc;
}
*/
	  

/*
 * The client callback packet structure
 *
 *  0    1    2   3   4   5   6   7
 *  ---------------------------------
 * |0xAB|0xCD|  Len  | 0 | 1 | 2 | 3 |
 *  ---------------------------------
 * | Header  |Pkt Len| Int ret. val. |
 *  2 * uchar  ushort       int
 */

int client_dump(int capture_retval, u_long client_ip)
{
	struct sockaddr_in clt_addr;
	int callback_sock, rtn;
	u_char callback_pkt[8];
	u_short pkt_len;
	
	memset(&clt_addr, 0, sizeof(struct sockaddr_in));
	clt_addr.sin_family = AF_INET;
	clt_addr.sin_addr.s_addr = client_ip;
	clt_addr.sin_port = htons(61373);
	
	if((callback_sock=socket(AF_INET, SOCK_DGRAM, 0))==-1)
	{
		perror("RPCAP_SERVER::callback socket error");
		return -1;
	}

	/* format the packet */
	pkt_len = 0x08;
	callback_pkt[0] = 0xAB;
	callback_pkt[1] = 0xCD;
	memcpy(&callback_pkt[2], &pkt_len, 2);           /* all numbers sent in host order */
	memcpy(&callback_pkt[4], &capture_retval, 4);

	if((rtn = sendto(callback_sock, (void *)callback_pkt, 8, 0,
		(struct sockaddr *)&clt_addr, sizeof(clt_addr))) == -1)
	{
		perror("RPCAP_SERVER::client_dump: send failed");
		return -1;
	}

	return rtn;
}
	


void sigpipe_hdlr(int spipe)
{
	fprintf(stderr, "\nSIGPIPE received: The connection to the client process appears to be broken.\n");
	fprintf(stderr, "RPCAP_SVC will exit now.  Please check the client process status and restart.\n");
	exit(1);
}


int capture_cleanup(int capture_return)
{
	conn_flag=2;
	(*cparams->callback)(NULL, NULL, NULL);
	conn_flag=0;            /* reset the system */
	return client_dump(capture_return, cparams->cap_client.cl_ip);
}


int * rpcap_end_capture_1_svc(void *null_param, struct svc_req *rqst)
{
	static int result;

	if(pthread_cancel(cap_thread))
	{
		perror("Capture thread termination error\n");
		result = -1;
	}
	else
		result = 0;
	capture_cleanup(0);
	return &result;
}
	
