/*******************************************************************************************/
/*                              file rpcap_client.c                                        */
/*                                                                                         */
/*           Copyright (C) S. Krishnan <sri_krishnan@yahoo.com>, 2002                      */
/*                                                                                         */
/*     client file for the rpcap Remote Packet Capture library                             */
/*     language: C                                                                         */
/*     Compiler: gcc                                                                       */
/*     Dependencies: RPC and pthread libraries                                             */
/*******************************************************************************************/
 
 
/*
 * This file is part of the RPCAP Remote Packet Capture System.
 *
 * RPCAP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* #include </usr/local/include/pcap.h> */


#ifdef SYS_LIBPCAP
    #include <pcap/pcap.h>
#elif defined LOCAL_LIBPCAP
    #include </usr/local/include/pcap.h>
#endif /* LIBPCAP includes */

#include <errno.h>
#include "pcap-int.h"
#include "rpcap.h"



/*---------------------------------Globals---------------------------------*/

#define OPEN 1
#define CLOSE 0
#define MAXCOUNT 8192

/* extern char *errbuf; */

pcap_t local_adapter;
char kerrbuf[256];
int initialized=0;
CLIENT *cl;
char server[256];
struct cl_addr in_data_channel;
struct sockaddr_in target;
volatile int clt_fd=0;
int capture_loop=0;
int cap_error=0;
struct data_channel
{
	char host_addr[20];
	u_short host_port;
} data_channel;

volatile int cap_fd;

pcap_handler user_callback;
u_char *user_buffer;

typedef enum cap_func_index
{
	PCAP_DISPATCH=0,
	PCAP_LOOP,
} CAP_FUNC_INDEX;

char dev_lbl[20];
char rpc_error[256];
/*-----------------------------------------------------------------------*/

void remote_capture();
pthread_t create_listen_thread();
int pcap_capture_callback();

/*-----------------------------------------------------------------------*/


/*
 * we use the same pcap.h file that libpcap uses, in order to maintain an identical API
 */

int init_rpcap(char *inserver, char *client_addr, u_short client_port)
{
	static struct timeval new_timeout = {140,0};
	struct hostent *client_host;
	
	if((cl = clnt_create(inserver, RPCAP, RPCAPVER, "tcp"))==NULL)
	{
		perror("INITIALIZE: error creating client\n");
		return 0;
	}
	clnt_control(cl, CLSET_TIMEOUT, (char *)&new_timeout);
	memset(server, 0, sizeof(char));
	capture_loop=0;
	cap_error=0;
	strcpy(data_channel.host_addr, client_addr);
	data_channel.host_port=client_port; /* 21373 nominal - note this is in host order */  
	in_data_channel.cl_ip = inet_addr(client_addr);
	/* inet_addr produces network order number */
	in_data_channel.cl_port = client_port; /* 21373 - host order, again */
	initialized=1;
	return initialized;
}


pcap_t *pcap_open_live(char *device, int snaplen, int promisc, int to_ms, char *ebuf)
{
	RPCAP_OPEN_LIVE_PARAMS open_params;
	adapter_handle *rem_adapter;

	if(!initialized)
	{
		perror("RPCAP: system not initialized\n");
		return NULL;
	}

	open_params.device=device;
	open_params.snaplen=snaplen;
	open_params.promisc=promisc;
	open_params.to_ms = to_ms;
	
	rem_adapter = *rpcap_open_live_1(&open_params, cl);
	if(rem_adapter==NULL)
	{
		fprintf(stderr, "pcap_open_live returned NULL\n");
		return NULL;
	}
	if((pcap_t *)rem_adapter->adapter_pcap.adapter_pcap_val==NULL)
	{
		return NULL;
	}
	return (pcap_t *)rem_adapter->adapter_pcap.adapter_pcap_val;

}


char *pcap_lookupdev(char *errbuf)
{
	RPCAP_FN_RTN_PTR rpcap_retval;

	rpcap_retval=*rpcap_lookupdev_1(NULL, cl);
	if(rpcap_retval->retval == -1)
	{
		strcpy(errbuf, rpcap_retval->ebuf);
		return NULL;
	}
	strcpy(dev_lbl, rpcap_retval->ebuf);
	return(&dev_lbl[0]);
}


int pcap_lookupnet(char *device, bpf_u_int32 *netp, bpf_u_int32 *maskp, char *errbuf)
{
	RPCAP_LOOKUPNET_PARAMS lookup_params;
	LOOKUPNET_RESULT_PTR lookup_result;
	
	lookup_result=*rpcap_lookupnet_1(&device, cl);
	*netp = lookup_result->netp;
	*maskp = lookup_result->maskp;
	errbuf = lookup_result->ebuf;
	return lookup_result->lookupnet_ret_val;
}


int set_indata_port(int setting)
{
	struct sockaddr_in clt;
	u_long addr;
	int addrlen;
	sigset_t sigmask;
	int ctl_mask;
	int on;
	int off=0;
	
	memset(&clt, 0, sizeof(struct sockaddr_in));
	if(setting==OPEN)
	{
		if((data_channel.host_port <=0) || !strlen(data_channel.host_addr))
		{
			perror("set_indata_port: invalid incoming socket parameters");
			return -1;
		}
		/* open the client socket */
		clt.sin_family=AF_INET;
		clt.sin_port=htons(data_channel.host_port);

		if((addr = inet_addr(data_channel.host_addr))==INADDR_NONE)
		{
			perror("set_indata_port: Client host address is invalid");
			return -1;
		}
		clt.sin_addr.s_addr = addr;
		if((clt_fd=socket(AF_INET, SOCK_STREAM, 0)) <0)
		{
			perror("set_indata_port: Error opening local data channel socket");
			return -1;
		}
		if(bind(clt_fd, (struct sockaddr *)&clt, sizeof(clt)) <0)
		{
			perror("set_indata_port: Unable to bind local data port");
			return -1;
		}
		return clt_fd;
	}
	
	if(setting==CLOSE)
		return close(clt_fd);

	return 0;
}


int pcap_capture(pcap_t *p, int cnt, pcap_handler callback, u_char *user, CAP_FUNC_INDEX capture_function)
{
	adapter_handle remote_adapter;
	RPCAP_CAPTURE_PARAMS capture_settings;
	int retval, cap_call_rtn, *ret_ptr;
	static struct timeval new_timeout = {140,0};
	pthread_t listen_tid;

	ret_ptr = &cap_call_rtn;
	
	// clnt_control(cl, CLSET_TIMEOUT, (char *)&new_timeout);
	user_callback = callback;
	user_buffer = user;
	
	if(set_indata_port(OPEN)==-1)
	{
		perror("Could not open incoming data socket from remote system");
		return -1;
	}
	
	listen_tid=create_listen_thread();
	
	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	
	capture_settings.p = &remote_adapter;
	capture_settings.cnt = cnt;
	capture_settings.client_address.cl_ip = in_data_channel.cl_ip;
	capture_settings.client_address.cl_port = in_data_channel.cl_port;
	capture_settings.cap_func_sel = (u_int)capture_function;
	
	printf("PCAP_CAPTURE: Starting pcap_capture\n");

	ret_ptr=rpcap_capture_1(&capture_settings, cl);

	if(ret_ptr==NULL)
	{
		fprintf(stderr, "RPCAP_CAPTURE_1 returned NULL\n");
		return -1;
	}
	if(*ret_ptr==-1)
	{
		fprintf(stderr, "Error starting remote capture operation\n");
	 	return -1;
	}

	retval=pcap_capture_callback();             /* block on the remote callback loop */

	capture_loop=0;
	set_indata_port(CLOSE);
	return retval;
	
}


int pcap_dispatch(pcap_t *p, int cnt, pcap_handler callback, u_char *user)
{
	return pcap_capture(p, cnt, callback, user, PCAP_DISPATCH);
}


int pcap_loop(pcap_t *p, int cnt, pcap_handler callback, u_char *user)
{
	return pcap_capture(p, cnt, callback, user, PCAP_LOOP);
}


int pcap_compile(pcap_t *p, struct bpf_program *fp, char *str, int optimize, bpf_u_int32 netmask)
{
	RPCAP_COMPILE_PARAMS compile_params;
	adapter_handle remote_adapter;
	bpf_prg **filter_program;
	int *retval;

	char *backup_str = "NULL";

	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	compile_params.p = &remote_adapter;
	/*
	 * check to see if str is NULL, since RPC does not allow null strings.  If so, place the string "NULL" in str
	 * and pass it to the server to decipher and pass a NULL to the real pcap_compile.  While it would have been
	 * nice to avoid passing NULL strings to pcap_compile, unfortunately this is exactly what tcpdump does - this
	 * kludge is to work around this
	 */
	if (str == NULL)
		compile_params.str = backup_str;
	else
		compile_params.str = str;
	compile_params.optimize = optimize;
	compile_params.netmask = netmask;
	
	filter_program = rpcap_compile_1(&compile_params, cl);
	if (filter_program != NULL)
	{
		memcpy(fp, (*filter_program)->bpf_prgrm.bpf_prgrm_val, sizeof(struct bpf_program));
		return 0;
	}
	return -1;                       /* if error */
}


int pcap_setfilter(pcap_t *p, struct bpf_program *fp)
{
	RPCAP_SETFILTER_PARAMS setfilter_params;
	adapter_handle remote_adapter;
	bpf_prg filter_program;
	int *retval;
	
	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	filter_program.bpf_prgrm.bpf_prgrm_len = sizeof(struct bpf_program);
	(struct bpf_program *)filter_program.bpf_prgrm.bpf_prgrm_val=fp;
	
	setfilter_params.p = &remote_adapter;
	setfilter_params.fp = &filter_program;

	retval = rpcap_setfilter_1(&setfilter_params, cl);
	if (retval == NULL)
		return -1;
	return *retval;
}


void pcap_close(pcap_t *p)
{
	adapter_handle remote_adapter;
	int *result;
	
	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	
	result=rpcap_close_1(&remote_adapter, cl);
	switch((long)result)
	{
		case NULL:
			goto close_error;
			break;
		default:
			if(*result != 1)
			{
				goto close_error;
			}
			else
				goto close_done;
			break;
	}
close_error:
	fprintf(stderr, "Error closing remote capture handle\n");
	clnt_perror(cl, rpc_error);
	fprintf(stderr, "RPC ERROR %s\n", rpc_error);
close_done:
	
}

int pcap_stats(pcap_t *p, struct pcap_stat *ps)  
{
	RPCAP_STATS_WRAPPER *stats;
	adapter_handle remote_adapter;

	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	
	if((stats = *rpcap_stats_1(&remote_adapter, cl))==NULL)
	{
		fprintf(stderr, "RPCAP: Error getting remote pcap_stats\n");
		return -1;
	}
	memcpy(ps, stats->pcap_stats.pcap_stats_val, sizeof(struct pcap_stat));
	return stats->stats_retval;
}



int pcap_datalink(pcap_t *p)
{
	adapter_handle remote_adapter;
	int retval, *rptr;
	
	rptr = &retval;
	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	
	rptr = rpcap_datalink_1(&remote_adapter, cl);
	if(rptr == NULL)
	{
		fprintf(stderr, "RPCAP::rpcap_datalink returned NULL\n");
		clnt_perror(cl, rpc_error);
		fprintf(stderr, "RPC ERROR %s\n", rpc_error);
	}
	return *rptr;
}



int pcap_snapshot(pcap_t *p)
{
	adapter_handle remote_adapter;
	int retval, *rptr;

	rptr = &retval;
	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;

	rpcap_snapshot_1(&remote_adapter, cl);
	rptr = rpcap_snapshot_1(&remote_adapter, cl);
	if(rptr==NULL)
	{
	fprintf(stderr, "pcap_snapshot returned NULL\n");
	/* fprintf(stderr, "ERROR:: %s\n", errbuf); */
	return -1;
	}
	return *rptr;
}



char *pcap_geterr(pcap_t *p)
{
	adapter_handle remote_adapter;
	RPCAP_FN_RTN_PTR *rpcap_retval;

	remote_adapter.adapter_pcap.adapter_pcap_len = sizeof(pcap_t);
	(pcap_t *)remote_adapter.adapter_pcap.adapter_pcap_val=p;
	
	rpcap_retval=rpcap_geterr_1(&remote_adapter, cl);
	if (rpcap_retval == NULL)
	{
		strcpy(kerrbuf, "No errors");
		return kerrbuf;
	}
	if((*rpcap_retval)->ebuf == NULL)
		strcpy(kerrbuf, "RPC ERROR");
	else strcpy(kerrbuf, (*rpcap_retval)->ebuf);
	return(kerrbuf);
}



void pcap_perror(pcap_t *p, char *prefix)
{
	fprintf(stderr, "%s: %s\n", prefix, pcap_geterr(p));
}


/* we're kludging pcap_strerror for the time being, since rpcap is (presently) released only for Linux */
/* this quick 'n' dirty fix is to present a consistent libpcap-compatible API for user programs */
/* the strerror call has to be on the remote (server) host */
/*
char *pcap_strerror(int error)
{
	return *rpcap_strerror_1(&error, cl);
}
*/

void remote_capture()
{
	void *buf_ptr;
	int maxread, readcount=0, cap_header_size, addrlen, pktlen=0;
	struct pcap_pkthdr cap_header;
	u_int sigmask, sigoldmask, cap;
	u_char sock_buf[16384];
	u_short hdrlen, hlen;
	long timebuf;
	
	cap_header_size=sizeof(struct pcap_pkthdr);
	
	if(listen(clt_fd, 5) <0)
	{
		perror("data_dumper: error on client listen\n");
		cap_error=1;
		capture_loop=0;
	}
		
	capture_loop=1;

	/* printf("Entered capture loop\n"); */
	addrlen = sizeof(target);
		
	if((cap_fd = accept(clt_fd, (struct sockaddr *)&target, &addrlen)) <0)
	{
		perror("remote_capture: error on socket accept");
		exit(0);
	}

	/* read the data and invoke the callback on it */
	while(1)
	{
		buf_ptr = &sock_buf[0];
		/* first, the header */
		/*
		  TODO: Set up a struct with the header stuff, and read the entire header
		  into it as a block, to save on some reads - maybe on revisit.
		*/
		memset(&cap_header, 0, cap_header_size);
		read(cap_fd, &hdrlen, sizeof(u_short));
		hlen = ntohs(sizeof(hdrlen));
		read(cap_fd, &timebuf, sizeof(long));
		cap_header.ts.tv_sec = ntohl(timebuf);
		read(cap_fd, &timebuf, sizeof(long));
		cap_header.ts.tv_usec = ntohl(timebuf);
		read(cap_fd, &cap, sizeof(unsigned));
		cap_header.caplen = ntohl(cap);
		read(cap_fd, &cap, sizeof(unsigned));
		cap_header.len = ntohl(cap);
		
		/* and next, the data */
		pktlen=cap_header.caplen;

		while(pktlen >0)
		{
			readcount = read(cap_fd, (void *)buf_ptr, cap_header.caplen);
			if(readcount <0)
			{
				perror("remote_capture: error reading from remote capture data socket");
				cap_error=1;
				capture_loop=0;
				break;
			}
			if(readcount==0)
				break;
			buf_ptr+=readcount;
			pktlen-=readcount;
		}

		if(pktlen <=0 || readcount == 0)        /* we have all the octets in this packet */
		{
			buf_ptr=&sock_buf[0];
			user_callback(user_buffer, &cap_header, (u_char *)buf_ptr);
		}
		if(!capture_loop)
		{
			printf("\nCLIENT: Capture ended\n");
			close(cap_fd);
			break;
		}
	}
}



pthread_t create_listen_thread()
{
	pthread_t listen_tid;
	
	if(pthread_create(&listen_tid, NULL, (void *)&remote_capture, NULL))
	{
		perror("Listen thread creation error");
		return -1;
	}
	/* printf("Listen thread created\n"); */
	return listen_tid;
}



/*
 * I originally implemented the callback as a conventional RPC callback function
 * with an invocation to the server side to register the callback function and
 * then entered svc_run() here to wait for the server's capture loop to return.
 * However, this seemed wasteful for a one-shot operation, so I did the UDP
 * socket thing below, which is much simpler and lower on overhead too.
 *
 *
 * The callback packet structure:
 *   0    1    2   3   4   5   6   7
 *  ---------------------------------
 * |0xAB|0xCD|  Len  | 0 | 1 | 2 | 3 |
 *  ---------------------------------
 * | Header  |Pkt Len| Int ret. val. |
 *  2 * uchar  ushort       int
 */

int pcap_capture_callback()
{
	// open a UDP socket
	int callback_socket, recd_bytes;
	socklen_t len;
	struct sockaddr_in callbk_sock, callback_from;
	int listen_loop=1;
	u_char callback_buf[10], *buf_ptr, first_byte, second_byte;
	u_short pkt_len, tmpshort;
	int retval, tmplong;
	
	buf_ptr = callback_buf;
	callbk_sock.sin_family = AF_INET;
	callbk_sock.sin_addr.s_addr = inet_addr(data_channel.host_addr);
	callbk_sock.sin_port = htons(61373);

	if((callback_socket = socket(AF_INET, SOCK_DGRAM, 0))==-1)
	{
		perror("RPCAP: Callback Socket");
		return -1;
	}

	if(bind(callback_socket, (struct sockaddr *)&callbk_sock, sizeof(callbk_sock)) <0)
	{
		perror("RPCAP Callback: Unable to bind local data port");
		return -1;
	}
	
	len = sizeof(callback_from);
	memset(callback_buf, 0, 10);
	while(listen_loop)
	{
		 recd_bytes = recvfrom(callback_socket,  (void  *)callback_buf, 8, 0,
			 (struct sockaddr *)&callback_from, &len);
		 if(recd_bytes <8)
			 continue;
		 if(*buf_ptr==0xAB && *(buf_ptr+1)==0xCD)
		 {
			 buf_ptr+=2;
			 if((pkt_len=*(u_short *)memcpy(&tmpshort, buf_ptr, 2))!=0x08 /* packet len */)	 
			 {
				 fprintf(stderr, "RPCAP::pcap_capture_callback: Error reading callback socket\n");
				 return -1;
			 }
			 buf_ptr+=2;
			 retval=*(int *)memcpy(&tmplong, buf_ptr, 4);
			 listen_loop=0;
			 capture_loop=0;
			 break;
		 }
	}
	return retval;
}


	
int rpcap_end_capture_handler(void)
{
	int *result;
	/* fprintf(stderr, "INVOKED CAPTURE HANDLER\n"); */
	result = rpcap_end_capture_1(NULL, cl);
	if ((result == NULL) || (*result != 0))
	{
		fprintf(stderr, "Error terminating remote capture operation\n");
		return -1;
	}
	capture_loop=0;
	set_indata_port(CLOSE);
	return *result;
}
